#include "serializers.h"

#include <cstdio>

#include "mesh.h"
#include "light.h"
#include "texture.h"
#include "phongmaterial.h"

namespace serializers
{

// STRUCTS

struct MeshHeader
{
    unsigned int flags;
    int nbPositions;
    int depth;
    int nbInstances_offsets;
    int nbIndices;
    int nameLength;
};

struct MaterialHeader
{
    glm::vec3 emission;
    glm::vec3 diffuse;
    glm::vec3 specular;
    float shininess;
};

struct LightHeader
{

};

// PRIVATE FUNCTIONS

bool serializeMesh(Mesh* mesh, const MeshHeader &header, FILE *file);

bool deserializeMesh(Mesh* mesh, const MeshHeader &header, FILE *file);

bool serializeMaterial(Material* material, FILE *file);

Material* deserializeMaterial(FILE *file);

template<typename T>
bool writeBuffer(const std::vector<T> &vec, std::FILE *file);

template<typename T>
bool readBuffer(std::vector<T> &vec, std::FILE *file);

// IMPLEMENTATIONS

template<typename T>
bool writeBuffer(const std::vector<T> &vec, std::FILE *file)
{
    size_t nbWritten = std::fwrite(vec.data(), sizeof(T), vec.size(), file);
    return (nbWritten == vec.size());
}

template<typename T>
bool readBuffer(std::vector<T> &vec, std::FILE *file)
{
    size_t nbRead = std::fread(vec.data(), sizeof(T), vec.size(), file);
    return (nbRead == vec.size());
}

bool saveMesh(const std::string &filename, Mesh *mesh)
{
    // open file
    std::FILE *file = std::fopen(filename.c_str(), "w");
    if(file == NULL)
        return false;
    
    // creating header
    MeshHeader header;
    header.flags = mesh->getFlags();
    if(header.flags & (1 << Mesh::MESH_3D))
        header.nbPositions = mesh->positions3D.size();
    else
    {
        mesh->positions2D.size();
        header.depth = mesh->getDepth();
    }
    header.nbIndices = mesh->indices.size();
    header.nbInstances_offsets = mesh->instances_offsets.size();
    header.nameLength = mesh->getName().size();
    
    // serializing
    bool ok = false;
    if(header.nbPositions != 0)
        ok = serializeMesh(mesh, header, file);
    
    // finishing
    std::fclose(file);
    return ok;
}

bool serializeMesh(Mesh* mesh, const MeshHeader &header, FILE *file)
{
    // writing header
    size_t nbWritten;
    
    nbWritten = std::fwrite(&header, sizeof(MeshHeader), 1, file);
    if(nbWritten != 1)
        return false;
    
    if(header.nameLength != 0)
    {
        nbWritten = std::fwrite(mesh->getName().data(), header.nameLength, 1, file);
        if(nbWritten != 1)
            return false;
    }

    // writing buffers
    if(header.flags & (1 << Mesh::MESH_3D))
    {
        if(!writeBuffer(mesh->positions3D, file))
            return false;
        if(mesh->normals.size() == 0)
            mesh->computeNormals();
        if(!writeBuffer(mesh->normals, file))
            return false;
        if(header.flags & (1 << Mesh::MESH_TANGENT_SPACE))
        {
            if(!writeBuffer(mesh->tangents, file))
                return false;
        }
    }
    else
    {
        if(!writeBuffer(mesh->positions2D, file))
            return false;
    }
    
    if(header.nbInstances_offsets)
    {
        if(!writeBuffer(mesh->instances_offsets, file))
            return false;
    }
    
    if(header.nbIndices)
    {
        if(!writeBuffer(mesh->indices, file))
            return false;
    }
    
    if(header.flags & (1 << Mesh::MESH_TEXTURABLE))
    {
        if(!writeBuffer(mesh->texCoords, file))
            return false;
    }
    
    return true;
}

bool serializeMaterial(Material* material, FILE *file)
{
    unsigned int flags = material->getFlags();
    
    if(!fwrite(&flags, sizeof(unsigned int), 1, file))
        return false;
    
    if(flags & (1 << Mesh::MATERIAL_PHONG))
    {
        MaterialHeader header;
        PhongMaterial *mat = (PhongMaterial*)material;
        header.emission = mat->emission;
        header.diffuse = mat->diffuse;
        header.specular = mat->specular;
        header.shininess = mat->shininess;
    }
    return true;
}

bool saveLight(const std::string &filename, Light *light)
{
    std::FILE *file = std::fopen(filename.c_str(), "w");
    if(file == NULL)
        return false;
    
    std::fclose(file);
}

Mesh* loadMesh(const std::string &filename)
{
    // open file
    std::FILE *file = std::fopen(filename.c_str(), "r");
    if(file == NULL)
        return NULL;
    
    // reading header
    MeshHeader header;
    size_t nbRead = std::fread(&header, sizeof(MeshHeader), 1, file);
    
    // deserializing mesh
    bool ok = false;
    Mesh *mesh = NULL;
    if(nbRead == 1 && header.nbPositions != 0)
    {
        mesh = new Mesh();
        ok = deserializeMesh(mesh, header, file);
    }
    
    // finishing
    std::fclose(file);
    
    if(!ok && mesh != NULL)
    {
        delete mesh;
        mesh = NULL;
    }
    return mesh;
}

bool deserializeMesh(Mesh* mesh, const MeshHeader &header, FILE *file)
{
    std::string name = "mesh";
    if(header.nameLength != 0)
    {
        name.resize(header.nameLength);
        if(!fread(&(name[0]), header.nameLength, 1, file))
            return false;
    }
    mesh->setName(name);
    
    if(header.flags & (1 << Mesh::MESH_3D))
    {
        mesh->positions3D.reserve(header.nbPositions);
        if(!readBuffer(mesh->positions3D, file))
            return false;
        mesh->normals.reserve(header.nbPositions);
        if(!readBuffer(mesh->normals, file))
            return false;
        if(header.flags & (1 << Mesh::MESH_TANGENT_SPACE))
        {
            mesh->tangents.reserve(header.nbPositions);
            if(!readBuffer(mesh->tangents, file))
                return false;
        }
    }
    else
    {
        mesh->positions2D.reserve(header.nbPositions);
        if(!readBuffer(mesh->positions2D, file))
            return false;
        mesh->setDepth(header.depth);
    }
    
    if(header.nbInstances_offsets)
    {
        mesh->instances_offsets.reserve(header.nbInstances_offsets);
        if(!readBuffer(mesh->instances_offsets, file))
            return false;
    }
    
    if(header.nbIndices)
    {
        mesh->indices.reserve(header.nbIndices);
        if(!readBuffer(mesh->indices, file))
            return false;
    }
    
    if(header.flags & (1 << Mesh::MESH_TEXTURABLE))
    {
        mesh->texCoords.reserve(header.nbPositions);
        if(!readBuffer(mesh->texCoords, file))
            return false;
    }
    
    mesh->setIsBillboard(header.flags & (1 << Mesh::MESH_BILLBOARD));
    mesh->setIsDoubleSided(header.flags & (1 << Mesh::MESH_DOUBLE_SIDED));
    mesh->setIsShadowCaster(header.flags & (1 << Mesh::MESH_SHADOWED));
    
    return true;
}


Material* deserializeMaterial(FILE *file)
{
    
}

Light* loadLight(const std::string &filename)
{

}

}
